import java.io.*;
import java.util.Scanner;

public final class FileReaderLbl {

 public FileReaderLbl(WorkerBee aBee) {
    Bee = aBee;
    fFile = new File( aBee.getInputFile() );
    lines = 0;
    counter = 0;
    offset = aBee.getInputFileOffset(); // where do I start reading in the file?
    testing = false;
    stop = false;
 }
  
  /** Template method that calls {@link #processLine(String)}.  */
  public void process() throws FileNotFoundException {
    Scanner scanner = new Scanner(fFile, Bee.getEncoding());
    try {
      //first use a Scanner to get each line - stop is for emergency stops
      while ( !stop && scanner.hasNextLine() ){
	if( Bee.shouldIstop() ) { Bee.emergencyStop(); } // emergencyStop throws exception that forces exit from this loop
        try { 
              if( processLine( scanner.nextLine() ) ) { Bee.receiveEOL(lines); } // don't send EOL signal for skipped lines
        } catch (Exception e) { System.err.println("inner process()"); e.printStackTrace(); }
      }
      Bee.receiveEOF(lines); // so that WorkerBee flushes its buffer to the database
    } catch (Exception e) { System.err.println("terminating process()..."); }
    finally {
      scanner.close(); //ensure the underlying stream is always closed
    }
  } // process() terminates and WorkerBee proceeds to unlock()

  /** 
  *
  * <P>This implementation expects a number followed by a colon and a space 
  * before the actual sql statements. </P>
  *
  */
  protected Boolean processLine(String aLine){
    // a simple way of implementing offset
    if( offset > lines ) { lines++; return false; } // false here means the line was completely skipped
    //use a second Scanner to parse the content of each line 
    Scanner scanner = new Scanner(aLine);
    scanner.useDelimiter(": ");
    if ( scanner.hasNext() ){
      String discarded = scanner.next();
      String sql = scanner.next();
      while( scanner.hasNext() ) { sql = sql + ": " + scanner.next(); } // just being careful
      if (!testing) {
          if (!this.Bee.receiveSql(sql.trim())) { emergencyStop(); scanner.close(); return false; } // avoid lines++
          else { counter++; }
      }
    }
    else {
      System.err.println("<"+Bee.getInputFile()+"> Line "+ (lines+1) +": Empty or invalid line. Unable to process.");
    }
    //(no need for finally here, since String is source)
    scanner.close();
    //
    lines++; // keeping track of line number for recovery
    return true; // true here means the line was properly processed - useful for offset
  }

  public void setTesting(Boolean val) { this.testing = val; }
  
  // PRIVATE //
  private final WorkerBee Bee;
  private File fFile;
  private File demoDataFile;
  private Boolean testing = false;
  private Integer counter; // this is counting lines passed back to WorkerBee
  private Integer lines; // this is counting total number of lines read
  private Integer offset; // this is number of lines skipped when restarting a stopped update
  private Boolean stop; // for emergency stops

  private void emergencyStop() {
     System.err.println("*** FileReaderLbl: An emergency stop has been requested!");
     Bee.receiveSqlError();
     stop = true;
  }
 
  private static void log(Object aObject){
    System.out.println(String.valueOf(aObject));
  }

}
 
