import java.io.*;
import java.sql.*;
import java.lang.*;
import java.util.Scanner;
import java.util.ArrayList;
import java.util.regex.*;
import org.w3c.dom.*;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException; 

public class WorkerBee extends Thread {

  public void setBuffer(Integer num) { SQL_BUFFER_NUMBER_OF_LINES = num; }

  public WorkerBee(pgBee queen, String inputfile) {

    master = queen;
    InputFile = inputfile;
    buffer = ""; // initializing buffer to empty string
    buffercount = 0;
    SQL_BUFFER_NUMBER_OF_LINES = 1; // set to a different value below from settings.xml
    started = false;
    stop = false;
    lines = 0; // last successful line read from InputFile - as given by the FileReaderLbl object
    counter = 0; // number of InputFile lines which were properly processed by this WorkerBee instance
    mylines = 0; // this is changed below once we try to read the TrackFile
    lastSuccess = 0; // ditto
    //
    // Reading configuration parameters from settings.xml
    try {
       DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
       DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
       settings = docBuilder.parse (new File("settings.xml"));
    }catch (SAXParseException err) {
         System.err.println ("*** settings.xml Parsing error" + ", line " 
              + err.getLineNumber () + ", uri " + err.getSystemId ());
         System.err.println(" " + err.getMessage ());
    }catch (SAXException e) {
         Exception x = e.getException ();
         ((x == null) ? e : x).printStackTrace ();
    }catch (Throwable t) {
         t.printStackTrace ();
    }

    encoding = pref("wb-file-encoding"); // we need this to successfully parse input files

    // ok, let's construct the jdbc connection string from vals in settings.xml
    String jdbcString ="jdbc:postgresql://"+pref("db-host")+":"+pref("db-port")+"/"+pref("db-database");

    // tracking progress line by line in InputFile is also quite important
    String[] temp = InputFile.split("/"); // getting the filename, with no path
    TrackFile = new File( pref("wb-track-dir") + "/" + temp[temp.length-1] );
    // code with readFileOffset() has been moved to the start() function - needs to occur after check for lock
    // ...
    // important for performance, read from settings.xml
    SQL_BUFFER_NUMBER_OF_LINES = Integer.parseInt(pref("wb-batch-size")); // set to a different value below from settings.xml
    // simple locking mechanism - to prevent conflict with other WorkerBee instances
    LockFile = new File( pref("wb-lock-dir") + "/" + temp[temp.length-1] );
    // for reporting purposes
    success = false; // this is changed to true when we receiveEOF()
    error = false; // this is changed when we receiveSqlError()

    // Database Connectivity
    try {
       Class.forName("org.postgresql.Driver");
    } catch (ClassNotFoundException cnfe) {
       System.err.println("*** Couldn't find the Postgresql JDBC driver!");
       System.exit(1);
    }
    connection = null;
    try {
       connection = DriverManager.getConnection(jdbcString,pref("db-username"),pref("db-password"));
    } catch (SQLException se) {
       System.err.println("*** Couldn't connect to  Postgresql server!");
       System.exit(1);
    }
    if (connection != null) {
       try { statement = connection.createStatement();
             System.out.println(this.toString()+ " database connection: OK");
       } catch (Exception e) {
             System.err.println("*** Couldn't create a JDBC statement!");
             System.exit(1);
       }
    }

  }

  public void run() {
      // If pgBee is in shutdown mode, there is no point even starting
      if ( master.shouldIstop() ) { return; }
      // Checking for the existence of a lockfile
      if (LockFile.exists()) {
          System.err.println("*** "+InputFile+" LOCKED by another instance or a stale lockfile");
          return;
      }
      else { // Good - It's ours for the taking!
          try { lock(); }
          catch (Exception e) { System.err.println("*** "+this+": Tried but failed to lock "+InputFile); return; }
      }
      started = true;
      // Now is the right time for reading the offset from TrackFile
      offset = readFileOffset(); // from the TrackFile
      mylines = offset; // mylines is for paranoid double-checking
      lastSuccess = offset-1; // in case we crash again from the very first batch
      try { updateTrackFile(lastSuccess); } // writing TrackFile at the very start, in case parallel==false
      catch (Exception e) { System.err.println("*** Updating the track file for "+InputFile+" has FAILED!"); e.printStackTrace(); System.exit(1); }
      // good: we've got a lock (and the offset) - we can start processing the (rest of the) file
      reader = new FileReaderLbl(this);
      reader.setTesting( master.getTesting() );
      try { reader.process(); } // unlocking has been moved to the housekeeping section of pgBee
      catch (Exception e) { System.err.println("*** From run(): "); e.printStackTrace(); receiveSqlError(); }
      return;
  }

  public Boolean shouldIstop() { return stop || master.shouldIstop() ; }

  public Boolean haveIstarted() { return started; }

  public String getInputFile() { return InputFile; }

  public Integer getInputFileOffset() { return offset; }

  public Boolean getSuccess() { return success; }

  public Boolean getError() { return error; }

  public Integer readFileOffset() {
      String myline = ""; Integer result = 0;
      Scanner myscanner;
      try {
            myscanner = new Scanner(TrackFile);
            if(myscanner.hasNextLine()) { myline = myscanner.nextLine().trim(); }
            myscanner.close();
      }
      catch (Exception e) { System.out.println ("Starting at the beginning :-) (no track file found) --"); myline = ""; }
      try { result = Integer.parseInt(myline); }
      catch (Exception e) { result = 0; }
      return result;
  }

  public String getEncoding() { return encoding; }

  public Boolean receiveSql(String sql) {
      buffer += sql; buffercount++;
      // we wait until we aggregate enough sql statements in the String buffer and then...
      if(SQL_BUFFER_NUMBER_OF_LINES <= buffercount) {
         try { processBufferWithTransaction(); } // we execute them in a single batch / transaction
         catch (Exception e) { e.printStackTrace(); stop = true; return false; }
         try { updateTrackFile(lines); } // if the transaction has gone well , we update the relevant track file
         catch (Exception e) { System.err.println("*** Updating the track file for "+InputFile+" has FAILED!"); e.printStackTrace(); System.exit(1); }
         lastSuccess = lines; // WorkerBee has not received the EOL yet
         try { Thread.sleep( master.sleepInterval() );  } // introducing some sleep, trying to be nice :-)
         catch (InterruptedException e ) { emergencyStop(); } // just in case we receive SIGTERM while sleeping
      }
      return true;
  }

  public void receiveEOL(Integer l) {
      lines = l; mylines++; counter++;
      if(mylines.equals(lines)) { // (mylines == lines) produced many errors [== tests object equality, not value]
          // we update the track directory once the database transaction has gone through, not here
          return;
      }
      else { System.err.println("*** mylines("+mylines+") != lines("+lines+") -> Problem with offset or something worse?"); }
      return;
  }

  public Boolean receiveEOF(Integer num) {
      try { processBufferWithTransaction(); }
      catch (Exception e) { stop = true; return false; } // perhaps it should do something more userful like a retry
      try { updateTrackFile(num); } // I have decided to trust the value received in receiveEOF()
      catch (Exception e) { System.err.println("*** Updating the track file for "+InputFile+" has FAILED!"); e.printStackTrace(); System.exit(1); }
      System.out.println(this+" has read "+counter+" lines from "+InputFile+" in this go.");
      success = true; // yes!
      return true;
  }

  public void emergencyStop() {
      try { updateTrackFile(lastSuccess); } // I have decided to trust the value received in receiveEOF()
      catch (Exception e) { System.err.println("*** Updating the track file for "+InputFile+" has FAILED!"); e.printStackTrace(); System.exit(1); }
      System.out.println(this+" has read "+ counter +" lines from "+InputFile+" in this go.");
      throw new RuntimeException("Emergency Stop");
  }

  public void receiveSqlError() {
      System.out.println("---> "+this+" has received an instruction to retry from line: "+ (lastSuccess+1) );
      error = true; // for indicating there was an error, not just an interrupt
      try { updateTrackFile(lastSuccess); } // if the transaction has gone well , we update the relevant track file
      catch (Exception e) { System.err.println("*** SqlError: updating track file for "+InputFile+" has FAILED!"); e.printStackTrace(); System.exit(1); }
      throw new RuntimeException("Emergency Stop"); // so that the thread does not terminate normally and TrackFile does not get overwritten
  }

  // called from pgBee, to avoid "unexpected EOF on client connection" in Postgres log files
  public void closeConnection() throws SQLException { connection.close(); }

  // PRIVATE //
  private pgBee master;
  private String InputFile;
  private File TrackFile;
  private File LockFile;
  private FileReaderLbl reader;
  private Connection connection;
  private Statement statement;
  private Document settings;
  private String buffer; // a string buffer, something like a bucket for ; separated sql statements
  private Boolean stop; // for emergency stops
  private Boolean started; // pgBee (in serial mode) uses this to check whether WorkerBee has started while cleaning up
  private Integer buffercount; // for keeping track of entries in the string buffer
  private Integer SQL_BUFFER_NUMBER_OF_LINES; // this may be changed for performance optimizations
  private Integer lines; // keeping track of last successful line read in InputFile
  private Integer offset; // line Offset from TrackFile
  private Integer mylines; // paranoid double-checking of lines
  private Integer counter; // number of lines actually processed by this WorkerBee
  private Integer lastSuccess; // last line number of the last successful batch
  private Boolean success; // for reporting back to WiDUp instance
  private Boolean error; // trying to differentiate between interruptions and errors
  private String encoding; // there were some problems with latin1 input files, so we're being explicit
 
  private void processBufferWithTransaction() throws Exception {
     try {
        connection.setAutoCommit(false);
        processBuffer();
        connection.commit();
     } catch (Exception ex) {
         try {
            connection.rollback();
            throw new Exception("*** Transaction failed: " + ex.getMessage());
         } catch (SQLException sqx) {
            throw new Exception("*** Rollback failed: " + sqx.getMessage());
         }
     } finally { buffer = ""; buffercount = 0; } // resetting buffer to empty string and buffercount to zero
  }

  private void processBuffer() throws Exception {
     statement.clearBatch();
     buffer = buffer.replaceAll("\\\\(.)", "'||E'\\\\$1'||'"); // turning old-style escapes into proper escape string syntax
     buffer = buffer.replaceAll("\'\\|\\|E\'\\\\\'\'\\|\\|\'", "\'\'"); // defending against C-style single-quote escapes \'
     String[] individual = mysplit(buffer,';'); // splitting buffer into individual statements
     for (int i=0; i < individual.length; i++) { statement.addBatch(individual[i]); } // there is often space after last semi-colon
     statement.executeBatch();
  }

  private byte[] fixPounds(byte[] c) {
     // char[] c = CharBuffer.wrap(input).array(); // input.toCharArray() is an alternative to this
     ArrayList<Byte> list = new ArrayList<Byte>(); ArrayList<Integer> toBeRemoved = new ArrayList<Integer>();
     for (int i = 0; i < c.length; i++) { list.add((Byte)c[i]); }
     for (int i = 0; i < list.size(); i++) { if ( c[i-1] == (byte)163 && c[i] == (byte)49 ) { toBeRemoved.add((Integer)i); } }
     for (int i = 0; i < toBeRemoved.size(); i++) { list.remove(toBeRemoved.get(i)); }
     byte[] newc = new byte[list.size()]; for (int i = 0; i < list.size(); i++) { newc[i] = (byte)list.get(i); }
     return newc;
  }

  private String[] mysplit(String mybuffer, char delimiter) {
     ArrayList<Integer> splitpoints = new ArrayList<Integer>();
     ArrayList<String> statements = new ArrayList<String>();
     char[] c = mybuffer.toCharArray();
     boolean text = false;
     for (int i = 0; i < c.length; i++) {
        if (c[i] == '\'') { text = !text; }
	if (c[i] == delimiter && !text) { splitpoints.add(i); }
     }
     int last = 0;
     for (int i = 0; i < splitpoints.size(); i++) {
        statements.add( mybuffer.substring( last,splitpoints.get(i) ) );
	last = splitpoints.get(i) + 1;
     }
     statements.add( mybuffer.substring(last) );
     return (String[])statements.toArray(new String[statements.size()]);
  }

  // just a helper method to retrieve values from the settings.xml file
  private String pref(String s) {
     try {
         NodeList myList = settings.getElementsByTagName(s);
         Element myElement = (Element)myList.item(0);
         NodeList myTexts = myElement.getChildNodes();
         return ((Node)myTexts.item(0)).getNodeValue().trim() ;
     } catch (Exception e) {
         System.err.println("*** You have not supplied all necessary data in settings.xml - Parameter list:");
         System.err.println("*** db-host, db-port, db-username, db-password, db-database,");
         System.err.println("*** wb-in-dir, wb-lock-dir, wb-track-dir, wb-log-dir, wb-rejects-dir, wb-done-dir,");
         System.err.println("*** wb-batch-size, wb-parallel, wb-number-of-threads, wb-file-encoding, wb-sleep,");
         System.err.println("*** global-lock");
         return "";
     }
   }

  protected void updateTrackFile(Integer line) throws FileNotFoundException, IOException {
    // if it's not there, it creates it
    if (!(TrackFile.createNewFile() || TrackFile.canWrite())) {
      throw new IllegalArgumentException("*** Cannot write to TrackFile: " + this.TrackFile);
    }
    //use buffering
    Writer output = new BufferedWriter(new FileWriter(this.TrackFile));
    try {
      // all we need to write is the last line from InputFile which was successfully processed PLUS 1
     output.write( "" + (line+1) +"\n" ); // ??? EOL signal comes later than receiveSql, so lines always falls behind
    }
    catch (Exception e) { System.err.println("*** Thought I could but I cannot write to TrackFile!"); System.exit(1); }
    finally {
      output.close();
    }
  }

  protected void deleteTrackFile() throws FileNotFoundException, IOException {
    if (!TrackFile.exists()) { return; }
    if (!TrackFile.canWrite()) {
      throw new IllegalArgumentException("*** Cannot delete TrackFile: " + this.TrackFile);
    }
    TrackFile.delete();
  }

  protected void approveInputFile() throws FileNotFoundException, IOException {
    File sourceFile = new File( InputFile );
    String[] temp = InputFile.split("/");
    File targetFile = new File( pref("wb-done-dir") + "/" + temp[temp.length-1] );
    if ( !sourceFile.canWrite() ) {
      throw new IllegalArgumentException("*** Cannot move InputFile: " + InputFile);
    }
    if ( targetFile.exists() ) { 
      if ( targetFile.canWrite() ) { targetFile.delete(); }
      else { throw new IllegalArgumentException("*** Cannot delete existing "+ InputFile +" in done/ directory"); }
    }
    copy( sourceFile, targetFile );
    sourceFile.delete();
  }

  protected void rejectInputFile() throws FileNotFoundException, IOException {
    Integer myoffset = readFileOffset();
    File sourceFile = new File( InputFile );
    String[] temp = InputFile.split("/");
    File targetFile = new File( pref("wb-rejects-dir") + "/" + temp[temp.length-1] );
    if ( !sourceFile.canWrite() ) {
      throw new IllegalArgumentException("*** Cannot move InputFile: " + InputFile);
    }
    if ( targetFile.exists() ) { 
      if ( targetFile.canWrite() ) { targetFile.delete(); }
      else { throw new IllegalArgumentException("*** Cannot delete existing "+ InputFile +" in rejects/ directory"); }
    }
    copy( sourceFile, targetFile );
    sourceFile.delete();
    File trackInfo = new File( pref("wb-rejects-dir") + "/" + temp[temp.length-1] + ".track" );
    if (!( trackInfo.createNewFile() || trackInfo.canWrite() )) {
      throw new IllegalArgumentException("*** Cannot create associated track file for: " + targetFile);
    }
    Writer output = new BufferedWriter(new FileWriter( trackInfo ));
    try { output.write( "" + myoffset.toString() +"\n" ); }
    catch (Exception e) { System.err.println("*** Thought I could but I cannot write to TrackFile!"); System.exit(1); }
    finally { output.close(); }
  }

  void copy(File src, File dst) throws IOException {
    InputStream in = new FileInputStream(src);
    OutputStream out = new FileOutputStream(dst);

    // Transfer bytes from in to out
    byte[] buf = new byte[1024];
    int len;
    while ((len = in.read(buf)) > 0) {
        out.write(buf, 0, len);
    }
    in.close();
    out.close();
  }

  // lock, unlock are now protected methods because we want pgBee to use them while housekeeping at the end
  protected void lock() throws FileNotFoundException, IOException {
    // if it's not there, it creates it
    if (!(LockFile.createNewFile() || LockFile.canWrite())) {
      throw new IllegalArgumentException("*** Cannot write to LockFile: " + this.LockFile);
    }
    //use buffering
    Writer output = new BufferedWriter(new FileWriter(this.LockFile));
    try {
      // writing the String representation of this instance - doesn't really matter
      output.write( "" + this.toString() +"\n" );
    }
    catch (Exception e) { System.err.println("*** Thought I could but I cannot write to LockFile!"); System.exit(1); }
    finally {
      output.close();
    }
  }

  protected void unlock() { LockFile.delete(); }

}

