import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException; 

public class pgBee {

    public static void main(String[] argv) {
       pgBee myself;
       System.out.println("--------- pgBee.java");
       if (argv.length > 3) {
          Boolean lock;
	  if (argv[0].equals("yes") || argv[0].equals("true")) { lock = true; } else { lock = false; }
	  Integer threads;
	  threads = Integer.parseInt(argv[1]);
	  Boolean parallel;
	  if (argv[2].equals("yes") || argv[2].equals("true")) { parallel = true; } else { parallel = false; }
	  Boolean test;
	  Integer sleep;
	  sleep = Integer.parseInt(argv[3]);
	  if (argv.length > 4) { test = true; System.out.println("This is a test run..."); }
	  else { test = false; }
	  myself = new pgBee(lock,threads,parallel,sleep,test);
	  myself.start();
       }
       else {
          System.out.println("pgBee USAGE: java ... pgBee <lock> <threads> <parallel> <sleep> [<test>]");
          System.out.println("e.g. :       java ... pgBee yes 15 yes 1000 test");
       }
    }

    public pgBee(Boolean lock, Integer threads, Boolean parallel, Integer sleep, Boolean test) {
	numBees = threads;
	useThreads = parallel;
	sleepInterval = sleep;
        testing = test;
	completed = false;
	stop = false;
        mybees = new Vector(); // for multiple WorkerBee threads
        //
        // Reading configuration parameters from settings.xml
        try {
           DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
           DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
           settings = docBuilder.parse (new File("settings.xml"));
        }catch (SAXParseException err) {
            System.err.println ("*** settings.xml Parsing error" + ", line " 
                  + err.getLineNumber () + ", uri " + err.getSystemId ());
             System.err.println(" " + err.getMessage ());
        }catch (SAXException e) {
             Exception x = e.getException ();
             ((x == null) ? e : x).printStackTrace ();
        }catch (Throwable t) {
             t.printStackTrace ();
        }
        // checking for the existence of a global pgBee lock file
        GlobalLock = new File( pref("global-lock") );
        if (GlobalLock.exists()) {
           System.out.println("found " + pref("global-lock") + " --> pgBee won't run!");
           System.exit(0);
        }
	if (lock) {
	   try {
	      // create global lock file
	      lock();
           } catch (Exception myex) {
              System.err.println("pgBee: Global Lock file creation FAILURE"); myex.printStackTrace();
              System.exit(1);
	   }
	}
	// defending against Ctrl-C, SIGINT etc.
        Runtime.getRuntime().addShutdownHook( new MyShutdownHook() );
    }

    public Boolean shouldIstop() { return stop; }

    public Integer sleepInterval() { return sleepInterval; }
    
    public void start() {

       // determining list of valid files which could be processed in this run
       File inDir = new File( pref("wb-in-dir") );
       String[] inFiles = inDir.list();
       java.util.Arrays.sort(inFiles);  // useful for determining precedence
       todo = new ArrayList<String>();
       File alock; Integer i;
       for(i=0;i<inFiles.length;i++) {
          alock = new File( pref("wb-lock-dir") + "/" + inFiles[i] );
	  if ( !alock.exists() ) { todo.add( pref("wb-in-dir") + "/" + inFiles[i] ); }
       }

       System.out.println("pgBee: I have found "+ todo.size() +" files to process (available threads: "+ numBees.toString() + ")");

       // assigning files to available WorkerBees
       WorkerBee one;
       for(i=0;i<numBees;i++) {
          if (todo.size() > 0) {
              one = new WorkerBee(this,todo.get(0));
              mybees.addElement(one);
              todo.remove(0); // taking this file off todo list of files
          }
       }

       if(useThreads) {
          for (Enumeration e = mybees.elements(); e.hasMoreElements();) {
              one = (WorkerBee)e.nextElement();
              one.start(); // hoping for multi-threaded heaven
          } }
       else {
           System.out.println("pgBee: You have disabled parallel mode!");
           for (Enumeration e = mybees.elements(); e.hasMoreElements();) {
	       one = (WorkerBee)e.nextElement();
               one.run(); // threads will be fired up sequentially
           }
       }

       completed = true; // so that the shutdown hook doesn't print EMERGENCY STOP right at the end of a successful run
       // Runtime.getRuntime().halt(0);

    }
   
    public void setTesting(Boolean val) { testing = val; }
    public Boolean getTesting() { return testing; }

    private Boolean testing;
    private Integer numBees;
    private Vector mybees;
    private Document settings;
    private Boolean useThreads;
    private Boolean completed;
    private Boolean stop;
    private ArrayList<String> todo;
    private File GlobalLock;
    private Integer sleepInterval;

    private class MyShutdownHook extends Thread {
       public void run() {
          stop = true; if (!completed) { System.err.println(">>>>>>>>> pgBee: EMERGENCY STOP!"); }
          Boolean threadsAreAlive = true; WorkerBee one;
          // waiting for threads to finish before halting the virtual machine
          while(threadsAreAlive) {
               threadsAreAlive = false; // but if a single thread is still running, this is set back to true
               for (Enumeration e = mybees.elements(); e.hasMoreElements();) {
                    one = (WorkerBee)e.nextElement();
                    if(one.isAlive()) { threadsAreAlive = true; } // checking whether they have finished or not
               }
          }
	  if (!testing) { // if this is a normal run
             for (Enumeration e = mybees.elements(); e.hasMoreElements();) {
                  one = (WorkerBee)e.nextElement();
                  if( one.getError() ) { // if there was an Sql Error --> rejects
	             System.out.println(one.getInputFile() + ": this file probably contains SQL errors");
		      try {
		         one.rejectInputFile();
		         one.deleteTrackFile();
			 one.closeConnection();
		         one.unlock();
		      } catch (Exception myex) {
		         System.err.println("pgBee: Housekeeping has FAILED! -- please inspect"); myex.printStackTrace();
		         System.exit(1);
		      }
	          }
	          else if ( one.getSuccess() ) { // if entire file was processed --> done
		      try {
		         one.approveInputFile();
		         one.deleteTrackFile();
			 one.closeConnection();
		         one.unlock();
		      } catch (Exception myex) {
		         System.err.println("pgBee: Housekeeping has FAILED! -- please inspect"); myex.printStackTrace();
		         System.exit(1);
		      }
	          }
	          else { // if there was an interruption --> file stays put
	              try {
			 one.closeConnection();
		         one.unlock();
		      } catch (Exception myex) {
		         System.err.println("pgBee: Housekeeping has FAILED! -- please inspect"); myex.printStackTrace();
		         System.exit(1);
		      }
	          }
             }
         }
	 else { // if this is just testing, we need to get rid of track & lock files but leave all else in their place
             for (Enumeration e = mybees.elements(); e.hasMoreElements();) {
                  one = (WorkerBee)e.nextElement();
		  try {
		     if ( one.haveIstarted() ) {
		        one.deleteTrackFile();
			one.closeConnection();
		        one.unlock();
                     }
		  } catch (Exception myex) {
		     System.err.println("pgBee: Housekeeping has FAILED! -- please inspect"); myex.printStackTrace();
		     System.exit(1);
		  }
             }

	 }
	 unlock(); // the global lock file is not needed any more
         System.out.println("<<< pgBee: shutdown sequence completed");
       }
    }
  
    private void lock() throws FileNotFoundException, IOException {
      // if it's not there, it creates it
      if (!(GlobalLock.createNewFile() || GlobalLock.canWrite())) {
        throw new IllegalArgumentException("*** Cannot write to GlobalLock: " + this.GlobalLock);
      }
      //use buffering
      Writer output = new BufferedWriter(new FileWriter(this.GlobalLock));
      try {
        // writing the String representation of this instance - doesn't really matter
        output.write( "" + this.toString() +"\n" );
      }
      catch (Exception e) { System.err.println("*** Thought I could but I cannot write to GlobalLock!"); System.exit(1); }
      finally {
        output.close();
      }
    }

    protected void unlock() { GlobalLock.delete(); }

    // just a helper method to retrieve values from the settings.xml file
    private String pref(String s) {
     try {
         NodeList myList = settings.getElementsByTagName(s);
         Element myElement = (Element)myList.item(0);
         NodeList myTexts = myElement.getChildNodes();
         return ((Node)myTexts.item(0)).getNodeValue().trim() ;
     } catch (Exception e) {
         System.err.println("*** You have not supplied all necessary data in settings.xml - Parameter list:");
         System.err.println("*** db-host, db-port, db-username, db-password, db-database,");
         System.err.println("*** wb-in-dir, wb-lock-dir, wb-track-dir, wb-log-dir, wb-rejects-dir, wb-done-dir,");
         System.err.println("*** wb-batch-size, wb-parallel, wb-number-of-threads, wb-file-encoding, wb-sleep,");
         System.err.println("*** global-lock");
         return "";
     }
   }

}

